from __future__ import annotations
from sympy.printing.pycode import PythonCodePrinter
from sympy.codegen.ast import Assignment
import sympy
from functools import partial

from ..ode import ODE
from .. import templates
from .base import CodeGenerator, Func, RHSArgument, SchemeArgument


# class GotranPythonCodePrinter(NumPyPrinter):
class GotranPythonCodePrinter(PythonCodePrinter):
    def _traverse_matrix_indices(self, mat):
        rows, cols = mat.shape
        return ((i, j) for i in range(rows) for j in range(cols))

    def _print_MatrixElement(self, expr):
        if expr.parent.shape[1] == 1:
            # Then this is a colum vector
            return f"{self._print(expr.parent)}[{expr.i}]"
        elif expr.parent.shape[0] == 1:
            # Then this is a row vector
            return f"{self._print(expr.parent)}[{expr.j}]"
        else:
            return super()._print_MatrixElement(expr)

    def _print_Float(self, flt):
        return self._print(str(float(flt)))

    def _print_Piecewise(self, expr):
        fst, snd = expr.args
        if isinstance(fst[0], Assignment):
            value = (
                f"{super()._print(fst[0].args[0])} = "
                f"{super()._print(fst[0].args[1])} if "
                f"{super()._print(fst[1])} else "
                f"{super()._print(snd[0].args[1])}"
            )
        else:
            value = super()._print_Piecewise(expr)

        return value


def squeeze_list(lst):
    if len(lst) == 1:
        return lst[0]
    return lst


def squeeze_list_of_lists(lst):
    return [squeeze_list(item) for item in lst]


class PythonCodeGenerator(CodeGenerator):
    def __init__(self, ode: ODE, apply_black: bool = True) -> None:
        super().__init__(ode)

        self._printer = GotranPythonCodePrinter()

        if apply_black:
            try:
                import black
            except ImportError:
                print("Cannot apply black, please install 'black'")
            else:
                # TODO: add options for black in Mode
                setattr(self, "_formatter", partial(black.format_str, mode=black.Mode()))

    @property
    def printer(self):
        return self._printer

    @property
    def template(self):
        return templates.python

    def _rhs_arguments(
        self,
        order: RHSArgument | str = RHSArgument.stp,
    ) -> Func:
        value = RHSArgument.get_value(order)

        argument_dict = {
            "s": "states",
            "t": "t",
            "p": "parameters",
        }

        argument_list = [argument_dict[v] for v in value]
        states = sympy.IndexedBase("states", shape=(self.ode.num_states,))
        parameters = sympy.IndexedBase("parameters", shape=(self.ode.num_parameters,))
        values = sympy.IndexedBase("values", shape=(self.ode.num_states,))

        return Func(
            arguments=argument_list,
            states=states,
            parameters=parameters,
            values=values,
            return_name="values",
            num_return_values=self.ode.num_states,
        )

    def _scheme_arguments(
        self,
        order: SchemeArgument | str = SchemeArgument.stdp,
    ) -> Func:
        value = SchemeArgument.get_value(order)

        argument_dict = {
            "s": "states",
            "t": "t",
            "d": "dt",
            "p": "parameters",
        }

        argument_list = [argument_dict[v] for v in value]
        states = sympy.IndexedBase("states", shape=(self.ode.num_states,))
        parameters = sympy.IndexedBase("parameters", shape=(self.ode.num_parameters,))
        values = sympy.IndexedBase("values", shape=(self.ode.num_states,))

        return Func(
            arguments=argument_list,
            states=states,
            parameters=parameters,
            values=values,
            return_name="values",
            num_return_values=self.ode.num_states,
        )



from __future__ import annotations

import lark
import sympy as sp
from . import sympytools
from .exceptions import InvalidTreeError

BINARY_OPERATIONS = {"add", "mul", "sub", "div", "pow"}


def relational_to_piecewise(expr: sp.Expr) -> sp.Piecewise:
    if expr.is_Relational:
        return sp.Piecewise(
            (1, expr),
            (0, True),
        )
    return expr


def binary_op(op: str, fst, snd):
    """Binary operation

    Parameters
    ----------
    op : str
        Operation to perform
    fst : sp.Expr
        First argument
    snd : sp.Expr
        Second argument

    Returns
    -------
    sp.Expr
        The result of the operation
    """
    fst = relational_to_piecewise(fst)
    snd = relational_to_piecewise(snd)

    if op == "add":
        return sp.Add(fst, snd, evaluate=False)
    if op == "sub":
        return sp.Add(fst, sp.Mul(sp.Integer(-1), snd, evaluate=False), evaluate=False)
    if op == "div":
        return sp.Mul(fst, sp.Pow(snd, sp.Integer(-1), evaluate=False), evaluate=False)
    if op == "mul":
        return sp.Mul(fst, snd, evaluate=False)
    if op == "pow":
        return sp.Pow(fst, snd, evaluate=False)

    raise RuntimeError(f"Invalid binary operation {op}")


def build_expression(
    root: lark.Tree,
    symbols: dict[str, sp.Symbol] | None = None,
) -> sp.Expr:
    """Build a sympy expression from a lark tree

    Parameters
    ----------
    root : lark.Tree
        The root of the tree
    symbols : dict[str, sp.Symbol], optional
        A dictionary with symbols, by default None

    Returns
    -------
    sp.Expr
        The sympy expression
    """
    symbols_: dict[str, sp.Symbol] = symbols or {}

    def expr2symbols(tree: lark.Tree):
        if tree.data == "variable":
            return symbols_[str(tree.children[0])]
        if tree.data == "scientific":
            return sp.sympify(tree.children[0])

        if tree.data == "signedatom":
            if tree.children[0] == "-":
                return -expr2symbols(tree.children[1])
            else:  # +
                return expr2symbols(tree.children[1])

        if tree.data in BINARY_OPERATIONS:
            return binary_op(
                tree.data,
                expr2symbols(tree.children[0]),
                expr2symbols(tree.children[1]),
            )

        if tree.data == "constant":
            if tree.children[0] == "pi":
                return sp.pi

        if tree.data == "func":
            # Children name (e.g 'log', 'exp', 'cos' etc) are methods
            # available in the sympy name space
            funcname = tree.children[0]
            if tree.children[0] == "abs":
                # Only exceptions is 'abs' which is 'Abs'
                funcname = "Abs"

            return getattr(sp, funcname)(expr2symbols(tree.children[1]))

        if tree.data == "logicalfunc":
            if tree.children[0] == "Conditional":
                return sympytools.Conditional(
                    cond=expr2symbols(tree.children[1]),
                    true_value=expr2symbols(tree.children[2]),
                    false_value=expr2symbols(tree.children[3]),
                )

            elif tree.children[0] == "ContinuousConditional":
                rel_op, arg1, arg2 = tree.children[1].children
                cond = sp.sympify(rel_op.value)(expr2symbols(arg1), expr2symbols(arg2))

                true_value = expr2symbols(tree.children[2])
                false_value = expr2symbols(tree.children[3])
                sigma = expr2symbols(tree.children[4])

                return sympytools.ContinuousConditional(
                    cond=cond,
                    true_value=true_value,
                    false_value=false_value,
                    sigma=sigma,
                )

            return getattr(sp, tree.children[0])(
                expr2symbols(tree.children[1]),
                expr2symbols(tree.children[2]),
            )

        raise InvalidTreeError(tree=tree)

    return expr2symbols(root)



from __future__ import annotations
import typing
import sympy

if typing.TYPE_CHECKING:  # pragma: no cover
    from .ode import ODE


def states_matrix(ode: "ODE") -> sympy.Matrix:
    return sympy.Matrix([state.symbol for state in ode.sorted_states()])


def rhs_matrix(ode: "ODE", max_tries: int = 20) -> sympy.Matrix:
    intermediates = {x.symbol: x.expr for x in ode.intermediates}
    rhs = sympy.Matrix([state.expr for state in ode.sorted_state_derivatives()])

    num_tries = 0
    while (any([rhs.has(k) for k in intermediates.keys()])) and num_tries < max_tries:
        rhs = rhs.xreplace(intermediates)
        num_tries += 1

    if num_tries == max_tries:
        raise RuntimeError("Maximum number of tries used")
    return rhs


def jacobi_matrix(ode: "ODE") -> sympy.Matrix:
    return rhs_matrix(ode).jacobian(states_matrix(ode))


def Conditional(cond, true_value, false_value):
    """
    Declares a conditional

    Arguments
    ---------
    cond : A conditional
        The conditional which should be evaluated
    true_value : Any model expression
        Model expression for a true evaluation of the conditional
    false_value : Any model expression
        Model expression for a false evaluation of the conditional
    """
    cond = sympy.sympify(cond)

    from sympy.core.relational import Relational
    from sympy.logic.boolalg import Boolean, BooleanFalse, BooleanTrue

    # If the conditional is a bool it is already evaluated
    if isinstance(cond, (BooleanFalse, BooleanTrue)):
        return true_value if cond else false_value

    if not isinstance(cond, (Relational, Boolean)):
        raise TypeError(
            "Cond %s is of type %s, but must be a Relational" " or Boolean." % (cond, type(cond)),
        )

    return sympy.functions.Piecewise(
        (true_value, cond),
        (false_value, sympy.sympify(True)),
        evaluate=True,
    )


def ContinuousConditional(cond, true_value, false_value, sigma=1.0):
    """
    Declares a continuous conditional. Instead of a either or result the
    true and false values are weighted with a sigmoidal function which
    either evaluates to 0 or 1 instead of the true or false.

    Arguments
    ---------
    cond : An InEquality conditional
        An InEquality conditional which should be evaluated
    true_value : Any model expression
        Model expression for a true evaluation of the conditional
    false_value : Any model expression
        Model expression for a false evaluation of the conditional
    sigma : float (optional)
        Determines the sharpness of the sigmoidal function
    """

    cond = sympy.sympify(cond)
    # FIXME: Use the rel_op for check, as some changes has been applied
    # FIXME: in latest sympy making comparision difficult
    if "<" not in cond.rel_op and ">" not in cond.rel_op:
        TypeError(
            "Expected a lesser or greater than relational for " "a continuous conditional .",
        )

    # Create Heaviside
    H = 1 / (1 + sympy.exp((cond.args[0] - cond.args[1]) / sigma))

    # Desides which should be weighted with 1 and 0
    if ">" in cond.rel_op:
        return true_value * (1 - H) + false_value * H

    return true_value * H + false_value * (1 - H)



import gotranx
import pytest
import sympy as sp
from gotranx.expressions import build_expression


@pytest.mark.parametrize(
    "expr, symbol_values, expected",
    [
        ("x = y + 1 ", {"y": 2}, 3),
        ("x = y - 1 ", {"y": 2}, 1),
        ("x = y * 2 ", {"y": 2}, 4),
        ("x = y / 2 ", {"y": 4}, 2),
        ("x = y * (a + b - c) / 2 ", {"y": 2, "a": 3, "b": 6, "c": 1}, 8),
    ],
)
def test_build_expression(expr, symbol_values, expected, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)

    symbols = {name: sp.Symbol(name) for name in symbol_values}
    value = result[0].value.tree
    sympy_expr = build_expression(value, symbols)
    assert sympy_expr.subs(symbol_values).evalf() == pytest.approx(expected)


def test_build_two_expressions(parser, trans):
    expr = """
    states(u=42)
    parameters(a=1, b=2)
    x = (1 + a) / b
    y = b  * x
    du_dt = y + x - u
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)
    t = sp.Symbol("t")

    old_ode = gotranx.ode.ODE(result.components)

    symbols = old_ode.symbols
    ode = gotranx.ode.ODE(
        gotranx.ode.resolve_expressions(result.components, symbols=symbols),
        t=t,
    )

    assert len(ode.components) == 1
    assert len(ode.components[0].parameters) == 2
    assert len(ode.components[0].states) == 1
    assert len(ode.components[0].assignments) == 3
    assert len(ode.components[0].intermediates) == 2
    assert len(ode.components[0].state_derivatives) == 1

    u = ode["u"].symbol
    a = ode["a"].symbol
    b = ode["b"].symbol
    x = ode["x"].symbol
    y = ode["y"].symbol

    assert ode["x"].expr.evalf() == ((a + 1) / b).evalf()
    assert ode["y"].expr.evalf() == (b * x).evalf()
    assert ode["du_dt"].expr.evalf() == (x + y - u).evalf()


def test_add_temporal_state(parser, trans):
    expr = """
    states(u=42)
    parameters(a=1, b=2)
    x = (1 + a) / b
    y = b  * x
    du_dt = y + x - u
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)
    t = sp.Symbol("t")
    components = gotranx.ode.add_temporal_state(components=result.components, t=t)
    ode = gotranx.ode.ODE(components=components, t=t)
    assert str(ode["u"].symbol) == "u(t)"
    assert str(ode["du_dt"].symbol) == "Derivative(u(t), t)"


def test_state_parameters_with_expression_values(parser, trans):
    expr = """
    states(u=2*4/3)
    parameters(a=3*5**2)
    x = (1 + a) / b
    y = b  * x
    du_dt = y + x - u
    """
    tree = parser.parse(expr)
    result = trans.transform(tree).components[0]

    assert result.find_state("u").value.args == (
        sp.Mul(2, 4, evaluate=False),
        sp.Pow(3, sp.Integer(-1), evaluate=False),
    )
    assert result.find_parameter("a").value == sp.Mul(
        3, sp.Pow(5, 2, evaluate=False), evaluate=False
    )
